/* OpenCP Module Player
 * copyright (c) '94-'04 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * Basic glue for the different console implementations for linux
 *
 * revision history: (please note changes here)
 *  -ss040614   Stian Skjelstad <stian@nixia.no>
 *    -first release
 */

#include "config.h"

#ifdef __linux
#include <linux/major.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "types.h"
#include "boot/plinkman.h"

#include "boot/lrmi-ocp.h"
#include "boot/console.h"
#include "poutput-curses.h"
#include "poutput-fb.h"
#include "poutput-dga.h"
#include "poutput-vcsa.h"
#include "boot/pmain.h"

static void reset_api(void);
static void (*console_clean)(void)=NULL;

static int console_init(void)
{
	struct stat st;
#ifdef __linux
	char _stdout[128];
	char _stdin[128];

	int test_vesa=0;
	int test_fb=0;
	int test_vcsa=0;
#endif
	int test_dga=0;

	reset_api();
	
	fprintf(stderr, "Initing console... ");
	fflush(stderr);
#ifdef __linux
	memset(_stdin, 0, sizeof(_stdin));
	memset(_stdout, 0, sizeof(_stdout));
	

	if (readlink("/proc/self/fd/0", _stdin, sizeof(_stdin)-1)<0)
	if (readlink("/dev/fd/0", _stdin, sizeof(_stdin)-1)<0)
	{
		fprintf(stderr, "Failed to read link /proc/self/fd/0\n");
		return -1;
	}
	if (readlink("/proc/self/fd/1", _stdout, sizeof(_stdout)-1)<0)
	if (readlink("/dev/fd/1", _stdout, sizeof(_stdout)-1)<0)
	{
		fprintf(stderr, "Failed to read link /proc/self/fd/1\n");
		return -1;
	}
	if (strcmp(_stdin, _stdout))
	{
		fprintf(stderr, "stdout and stdin does not come from the same device\n");
		return -1;
	}
	
	if (stat(_stdin, &st))
	{
		fprintf(stderr, "stat failed on %s\n", _stdin);
		return -1;
	}

	if ((st.st_mode&S_IFMT)!=S_IFCHR)
	{
		fprintf(stderr, "stdout/stdin is not a character device\n");
		return -1;
	}

	switch ((st.st_rdev&0xff00)>>8)
	{
		case TTY_MAJOR:
			fprintf(stderr, "We have a tty, testing:\n    VESA\n    Framebuffer (/dev/fb)\n    VCSA (/dev/vcsa)\n    Curses\n");
			test_vesa=1;
			test_fb=1;
			test_vcsa=1;
			break;
		case UNIX98_PTY_SLAVE_MAJOR:
#if (UNIX98_PTY_MAJOR_COUNT>=2)
		case UNIX98_PTY_SLAVE_MAJOR+1:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=3)
		case UNIX98_PTY_SLAVE_MAJOR+2:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=4)
		case UNIX98_PTY_SLAVE_MAJOR+3:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=5)
		case UNIX98_PTY_SLAVE_MAJOR+4:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=6)
		case UNIX98_PTY_SLAVE_MAJOR+5:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=7)
		case UNIX98_PTY_SLAVE_MAJOR+6:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=8)
		case UNIX98_PTY_SLAVE_MAJOR+7:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=9)
		case UNIX98_PTY_SLAVE_MAJOR+8:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=10)
		case UNIX98_PTY_SLAVE_MAJOR+9:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=11)
		case UNIX98_PTY_SLAVE_MAJOR+10:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=12)
		case UNIX98_PTY_SLAVE_MAJOR+11:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=13)
		case UNIX98_PTY_SLAVE_MAJOR+12:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=14)
		case UNIX98_PTY_SLAVE_MAJOR+13:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=15)
		case UNIX98_PTY_SLAVE_MAJOR+14:
#endif
#if (UNIX98_PTY_MAJOR_COUNT>=16)
		case UNIX98_PTY_SLAVE_MAJOR+15:
#endif
			fprintf(stderr, "We have a PTY, so using curses and testing for DGA\n");
			test_dga=1;
			break;
		default:
			fprintf(stderr, "Unknown console type, but we can still try curses and test for DGA\n");
			test_dga=1;
			break;
	}

	if (test_vesa)
	{
		if (!lrmi_is_available)
			fprintf(stderr, "VESA requires LRMI to function\n");
		else
			fprintf(stderr, "VESA: todo\n");
	}
	if (test_vcsa)
	{
		if (!vcsa_init(st.st_rdev&0xff))
		{
			console_clean=vcsa_done;
			if (test_fb)
				fb_init(st.st_rdev&0xff);
			return 0;
		}
	}
#else
	test_dga=1;
#endif

	if (!curses_init())
	{
		console_clean=curses_done;
		if (test_dga)
			dga_init(st.st_rdev&0xff);
		return 0;
	}

	return -1;
}

static void console_done(void)
{
	if (console_clean)
	{
		console_clean();
		console_clean=NULL;
	}
	reset_api();
}

static void __plSetTextMode(unsigned char x)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "plSetTextMode not implemented in this console driver\n");
#endif
}
static void __plSetBarFont(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "plSetBarFont not implemented in this console driver\n");
#endif
}
static void __displaystr(unsigned short y, unsigned short x, unsigned char attr, const char *str, unsigned short len)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "displaystr not implemented in this console driver\n");
#endif
}
static void __displaystrattr(unsigned short y, unsigned short x, const unsigned short *buf, unsigned short len)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "displaystrattr not implemented in this console driver\n");
#endif
}
static void __displaystrattrdi(unsigned short y, unsigned short x, const char *txt, const char *attr, unsigned short len)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "displaystrattrdi not implemented in this console driver\n");
#endif
}
static void __displayvoid(unsigned short y, unsigned short x, unsigned short len)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "displayvoid not implemented in this console driver\n");
#endif
}

static int __plSetGraphMode(int size)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "plSetGraphMode not implemented in this console driver\n");
#endif
	return -1;
}
static int  __plSetGraphPage(unsigned char x)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "plSetGraphPage not implemented in this console driver\n");
#endif
	return -1;
}
static void __gdrawchar(unsigned short x, unsigned short y, unsigned char c, unsigned char f, unsigned char b)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gdrawchar not implemented in this console driver\n");
#endif
}
static void __gdrawchart(unsigned short x, unsigned short y, unsigned char c, unsigned char f)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gdrawchart not implemented in this console driver\n");
#endif
}
static void __gdrawcharp(unsigned short x, unsigned short y, unsigned char c, unsigned char f, void *picp)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gdrawcharp not implemented in this console driver\n");
#endif
}
static void __gdrawchar8(unsigned short x, unsigned short y, unsigned char c, unsigned char f, unsigned char b)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gdrawchar8 not implemented in this console driver\n");
#endif
}
static void __gdrawchar8t(unsigned short x, unsigned short y, unsigned char c, unsigned char f)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gdrawchar8t not implemented in this console driver\n");
#endif
}
static void __gdrawchar8p(unsigned short x, unsigned short y, unsigned char c, unsigned char f, void *picp)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gdrawchar8p not implemented in this console driver\n");
#endif
}	
static void __gdrawstr(unsigned short y, unsigned short x, const char *s, unsigned short len, unsigned char f, unsigned char b)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gdrawstr not implemented in this console driver\n");
#endif
}	
static void __gupdatestr(unsigned short y, unsigned short x, const uint16_t *str, unsigned short len, uint16_t *old)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gupdatestr not implemented in this console driver\n");
#endif
}
static void __gupdatepal(unsigned char color, unsigned char red, unsigned char green, unsigned char blue)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gupdatepal not implemented in this console driver\n");
#endif
}
static void __gflushpal(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "gflushpal not implemented in this console driver\n");
#endif
}

static int __ekbhit(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "ekbhit not implemented in this console driver\n");
#endif
	return 0;
}
static int __egetch(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "egetch not implemented in this console driver\n");
#endif
	return 0;
}

static void __drawbar(unsigned short x, unsigned short yb, unsigned short yh, unsigned long hgt, unsigned long c)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "drawbar not implemented in this console driver\n");
#endif
}

static void __idrawbar(unsigned short x, unsigned short yb, unsigned short yh, unsigned long hgt, unsigned long c)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "idrawbar not implemented in this console driver\n");
#endif
}

static void __Screenshot(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "Screenshot not implemented in this console driver\n");
#endif
}
static void __TextScreenshot(int scrType)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "TextScreenshot not implemented in this console driver\n");
#endif
}
static void __RefreshScreen(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "RefreshScreen not implemented in this console driver\n");
#endif
}
static void __setcur(unsigned char y, unsigned char x)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "setcur not implemented in this console driver\n");
#endif
}
static void __setcurshape(unsigned short shape)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "setcurshape not implemented in this console driver\n");
#endif
}
static int __conRestore(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "conRestore not implemented in this console driver\n");
#endif
	return 1;
}
static void __conSave(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "conSave not implemented in this console driver\n");
#endif
}
static void __plDosShell(void)
{
#ifdef CONSOLE_DEBUG
	fprintf(stderr, "plDosShell not implemented in this console driver\n");
#endif
}

static void reset_api(void)
{
	_plSetTextMode=__plSetTextMode;
	_plSetBarFont=__plSetBarFont;
	_displaystr=__displaystr;
	_displaystrattr=__displaystrattr;
	_displaystrattrdi=__displaystrattrdi;
	_displayvoid=__displayvoid;

	_plSetGraphMode=__plSetGraphMode;
	_plSetGraphPage=__plSetGraphPage;
	_gdrawchar=__gdrawchar;
	_gdrawchart=__gdrawchart;
	_gdrawcharp=__gdrawcharp;
	_gdrawchar8=__gdrawchar8;
	_gdrawchar8t=__gdrawchar8t;
	_gdrawchar8p=__gdrawchar8p;
	_gdrawstr=__gdrawstr;
	_gupdatestr=__gupdatestr;
	_gupdatepal=__gupdatepal;
	_gflushpal=__gflushpal;

	_ekbhit=__ekbhit;
	_egetch=__egetch;

	_drawbar=__drawbar;
	_idrawbar=__idrawbar;

	_Screenshot=__Screenshot;
	_TextScreenshot=__TextScreenshot;
	_RefreshScreen=__RefreshScreen;
	_setcur=__setcur;
	_setcurshape=__setcurshape;

	_conRestore=__conRestore;
	_conSave=__conSave;

	_plDosShell=__plDosShell;
}

struct initcloseregstruct consoleReg={console_init, console_done};

char *dllinfo = "initcloseafter consoleReg";
struct linkinfostruct dllextinfo = {"poutput", "OpenCP Output Routines (c) 1994-04 Niklas Beisert, Tammo Hinrichs, Stian Skjelstad", DLLVERSION, 0};
