/* OpenCP Module Player
 * copyright (c) '94-'04 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * Curses console driver
 *
 * revision history: (please note changes here)
 *  -ss040613   Stian Skjelstad <stian@nixia.no>
 *    -first release
 *  -ss040902   Stian Skjelstad <stian@nixia.no>
 *    -added setcur, setcurshape and plDosShell
 *    -made displaystrattr not use bugfix for first blank cell (to avoid black
 *     cursor)
 *  -doj040907  Dirk Jagdmann  <doj@cubic.org>
 *    -enchanced some of the entries in the chr_table
 *  -ss040909   Stian Skjelstad <stian@nixia.no>
 *    -Made NCURSES_BUGFIX1 a config option instead: [curses] fixbadgraphic
 *  -ss040918   Stian Skjelstad <stian@nixia.no>
 *    -Tweaked conRestore, conSave and curses_init, so we get the bahavior we
 *     expect
 *  -ss040919   Stian Skjelstad <stian@nixia.no>
 *    -New logic for setcurshape
 */
#define _CONSOLE_DRIVER

#include "config.h"

#include <curses.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <unistd.h>
#include "types.h"

#include "poutput-curses.h"
#include "boot/console.h"
#include "stuff/poutput.h"
#include "boot/psetting.h"

#if defined(SIGWINCH) && defined(TIOCGWINSZ) && HAVE_RESIZETERM
#define CAN_RESIZE 1
#else
#define CAN_RESIZE 0
#endif

#if CAN_RESIZE
static void adjust(int sig);
#endif

static chtype attr_table[256];
static chtype chr_table[256];

static int Width, Height;

static int fixbadgraphic;

static void displayvoid(unsigned short y, unsigned short x, unsigned short len)
{
	move(y, x);
	while(len)
	{
		chtype output;
		output='X'|attr_table[plpalette[0]];

		addch(output);
		len--;
	}
}

static void displaystr(unsigned short y, unsigned short x, unsigned char attr, const char *buf, unsigned short len)
{
	move(y, x);
	while(len)
	{
		unsigned char ch=(*buf)&0xff;
		chtype output;
		if (((!ch)||(ch==' '))&&(!(attr&0x80))&&fixbadgraphic)
		{
			output=chr_table['X']|attr_table[plpalette[(attr&0xf0)+((attr>>4)&0xf)]];
			addch(output);
		} else {
			output=chr_table[ch]|attr_table[plpalette[attr]];
			addch(output);
		}
	
		if (*buf)
			buf++;
		len--;
	}
}


static void displaystrattr(unsigned short y, unsigned short x, const uint16_t *buf, unsigned short len)
{
	int first=1; /* we need this since we sometimes place the cursor at empty spots... dirty */

	move(y, x);
	while(len)
	{
		unsigned char ch=(*buf)&0xff;
		unsigned char attr=((*buf)>>8);
		chtype output;

		if (((!ch)||(ch==' '))&&(!(attr&0x80))&&fixbadgraphic)
		{
			if (first)
			{
				output=chr_table[ch]|attr_table[plpalette[attr]];
				first=0;
			} else
				output=chr_table['X']|attr_table[plpalette[(attr&0xf0)+((attr>>4)&0xf)]];
			addch(output);
		} else {
			first=1;
			output=chr_table[ch]|attr_table[plpalette[attr]];
			addch(output);
		}
		buf++;
		len--;
	}
}



static void vgaMakePal(void)
{
	int pal[16];
	char palstr[1024];
	int bg,fg;
	char scol[4];
	char const *ps2=NULL;

	strcpy(palstr,cfGetProfileString2(cfScreenSec, "screen", "palette", "0 1 2 3 4 5 6 7 8 9 A B C D E F"));

	for (bg=0; bg<16; bg++)
		pal[bg]=bg;

	bg=0;
	ps2=palstr;
	while (cfGetSpaceListEntry(scol, &ps2, 2) && bg<16)
		pal[bg++]=strtol(scol,0,16)&0x0f;

	for (bg=0; bg<16; bg++)
		for (fg=0; fg<16; fg++)
			plpalette[16*bg+fg]=16*pal[bg]+pal[fg];
}

static unsigned char bartops[18]="  ___...---===**#";
static unsigned char ibartops[18]="  ```^^^~~~===**#";
static void idrawbar(unsigned short x, unsigned short y, unsigned short height, unsigned long value, unsigned long c)
{
	char buf[60];
	int16_t i;
	short yh1=(height+2)/3;
	short yh2=(height+yh1+1)/2;

	if (value>((height*16)-4))
		value=(height*16)-4;

	for (i=0; i<height; i++)
	{
		if (value>=16)
		{
			buf[i]=ibartops[16];
			value-=16;
		} else {
			buf[i]=ibartops[value];
			value=0;
		}
	}
	y-=height-1;
	for (i=0; i<yh1; i++)
	{
		displaystr(y, x, c&0xff, buf+i, 1);
		y++;
	}
	c>>=8;
	for (i=yh1; i<yh2; i++)
	{
		displaystr(y, x, c&0xff, buf+i, 1);
		y++;
	}
	c>>=8;
	for (i=yh2; i<height; i++)
	{
		displaystr(y, x, c&0xff, buf+i, 1);
		y++;
	}
}
static void drawbar(unsigned short x, unsigned short y, unsigned short height, unsigned long value, unsigned long c)
{
	char buf[60];
	short i;
	short yh1=(height+2)/3;
	short yh2=(height+yh1+1)/2;

	if (value>((height*16)-4))
		value=(height*16)-4;

	for (i=0; i<height; i++)
	{
		if (value>=16)
		{
			buf[i]=bartops[16];
			value-=16;
		} else {
			buf[i]=bartops[value];
			value=0;
		}
	}
	for (i=0; i<yh1; i++)
	{
		displaystr(y, x, c&0xff, buf+i, 1);
		y--;
	}
	c>>=8;
	for (i=yh1; i<yh2; i++)
	{
		displaystr(y, x, c&0xff, buf+i, 1);
		y--;
	}
	c>>=8;
	for (i=yh2; i<height; i++)
	{
		displaystr(y, x, c&0xff, buf+i, 1);
		y--;
	}
}

static void plSetTextMode(unsigned char x)
{
	int i;

	_plSetGraphMode(-1);

	plScrHeight=Height;
	plScrWidth=Width;
	plScrMode=0;

	for (i=0;i<plScrHeight;i++)
		displayvoid(i, 0, plScrWidth);
}



#if CAN_RESIZE
int resized=0;
static void adjust(int sig)
{
	resized=1;
	signal(SIGWINCH, adjust);    /* some systems need this */
}
static void do_resize(void)
{
	struct winsize size;

	if (ioctl(fileno(stdout), TIOCGWINSZ, &size) == 0)
	{
		resize_term(size.ws_row, size.ws_col);
		wrefresh(curscr);   /* Linux needs this */

		Height=plScrHeight=size.ws_row;
		if ((Width=plScrWidth=size.ws_col)>CONSOLE_MAX_X)
			Width=plScrWidth=CONSOLE_MAX_X;
		else if (plScrWidth<80)
			Width=plScrWidth=80; /* If a console gets smaller than THIS, the user deserves to get fucked up his ASS */
	}
	resized=0;
}
#endif /* CAN_RESIZE */

static void RefreshScreen(void)
{
#if CAN_RESIZE
	if (resized)
		do_resize();
#endif
 	refresh();
}

static int buffer=ERR;

static int ekbhit(void)
{
	if (buffer!=ERR)
		return 1;
	buffer=getch();
	if (buffer!=ERR)
		return 1;
	RefreshScreen();
	return 0;
}

static int egetch(void)
{
	int retval;
	RefreshScreen();
	if (buffer!=ERR)
	{
		retval=buffer;
		buffer=ERR;
		return retval;
	}
	retval=getch();
	if (retval==ERR)
		retval=0;
	return retval;
}

static void setcur(unsigned char y, unsigned char x)
{
	move(y, x);
}

static void setcurshape(unsigned short shape)
{
	curs_set(!!shape);
}

static int conactive=0;

static int conRestore(void)
{
	if (!conactive)
		return 0;
	endwin();
	conactive=0;
	return 0;
}

static void conSave(void)
{
	if (conactive)
		return;
	fflush(stderr);
	refresh();
	cbreak();
	nodelay(stdscr, TRUE);
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	idlok(stdscr, FALSE);
	start_color();
	conactive=1;
}

static void plDosShell(void)
{
	pid_t child;
	if (!(child=fork()))
	{
		char *shell=getenv("SHELL");
		if (!shell)
			shell="/bin/sh";
		if (!isatty(2))
		{
			close(2);
			dup(1);
		}
		execl(shell, shell, NULL);
		perror("execl()");
		exit(-1);
	} else if (child>0)
	{
		while(1)
		{
			int status, retval;
			if ((retval=waitpid(child, &status, 0))<0)
			{
				if (errno==EINTR)
					continue;
			}
			break;
		}
	}
}

int curses_init(void)
{
	int i;

	fprintf(stderr, "Initing curses... (%s)\n",  curses_version());
	if ((fixbadgraphic=cfGetProfileBool("curses", "fixbadgraphic", 0, 0)))
		fprintf(stderr, "curses: fixbadgraphic is enabled in config\n");

	if (!initscr())
	{
		fprintf(stderr, "curses failed to init\n");
		return -1;
	}
	conSave();
#if CAN_RESIZE
	signal(SIGWINCH, adjust);        /* arrange interrupts to resize */
#endif
	_displayvoid=displayvoid;
	_displaystrattr=displaystrattr;
	_displaystr=displaystr;
	_RefreshScreen=RefreshScreen;
	___setup_key(ekbhit, egetch); /* filters in more keys */
	_plSetTextMode=plSetTextMode;
	_drawbar=drawbar;
	_idrawbar=idrawbar;

	_conRestore=conRestore;
	_conSave=conSave;
	_plDosShell=plDosShell;

	_setcur=setcur;
	_setcurshape=setcurshape;

	start_color();

	/* the attr_table is insane... on a regular i386 ncurses, you have 64 pairs, and entry 0 is locked to grey text on black, and colors apear in wrong order compared to DOS etc, so this is just butt fucking ugly, but it seems to do the work. The code in displayattr etc looks nice then, and fast enough */
	for (i=1;i<COLOR_PAIRS;i++)
	{
		unsigned char color_table[8] = {COLOR_BLACK, COLOR_BLUE, COLOR_GREEN, COLOR_CYAN, COLOR_RED, COLOR_MAGENTA, COLOR_YELLOW, COLOR_WHITE};
		unsigned char ch, fg, bg;

		ch=i^0x07;
		fg=color_table[ch&0x7];
		bg=color_table[(ch&0x38)>>3];
		init_pair(i, fg, bg);
	}
	for (i=0;i<256;i++)
	{
		attr_table[i]=COLOR_PAIR(((i&0x7)^0x07)+((i&0x70)>>1));

		if (i&0x08)
			attr_table[i]|=A_BOLD;
		if (i&0x80)
			attr_table[i]|=A_STANDOUT;

		chr_table[i]=i;
		if (i<32)
			chr_table[i]=i+32;
	}
	
	chr_table[0x00]=' ';
	chr_table[0x01]='S'; /* Surround */
	chr_table[0x04]=ACS_DIAMOND; /* looks good on the header line */
	chr_table[0x08]='?'; /* ??? */
	chr_table[0x09]='?'; /* ??? */

	chr_table[0x0a]='@'; /* when the heck does this occure */
	chr_table[0x07]='@'; /* and this?*/

	chr_table[0x0d]='@'; /* we want a note */
	chr_table[0x10]=ACS_RARROW;
	chr_table[0x11]=ACS_LARROW;
	chr_table[0x12]=ACS_PLMINUS; /* we want an up+down arrow */
/*	chr_table[0x12]='&'; *//* retrigger */
	chr_table[0x18]=ACS_UARROW;
	chr_table[0x19]=ACS_DARROW;
	chr_table[0x1a]='`'; /* pitch? */
	chr_table[0x1b]='\''; /* pitch? */
	chr_table[0x1d]=ACS_PLUS; /* speed pitch lock */
	chr_table[(unsigned char)0x81]='u'; /* ?? */
	chr_table[(unsigned char)0xb3]=ACS_VLINE;
	chr_table[(unsigned char)0xba]=ACS_VLINE;
	chr_table[(unsigned char)0xc0]=ACS_LLCORNER;
	chr_table[(unsigned char)0xc1]=ACS_BTEE;
	chr_table[(unsigned char)0xc2]=ACS_TTEE;
	chr_table[(unsigned char)0xc3]=ACS_LTEE;
	chr_table[(unsigned char)0xc4]=ACS_HLINE;
	chr_table[(unsigned char)0xda]=ACS_ULCORNER;
	chr_table[(unsigned char)0xdd]='#';
	chr_table[(unsigned char)0xf0]='#'; /* mid char of long peak power level */
	chr_table[(unsigned char)0xfa]=ACS_BULLET;
	chr_table[(unsigned char)0xf9]=ACS_BULLET/*|A_BOLD*/; /* TODO */
	chr_table[(unsigned char)0xfe]=ACS_BLOCK; /* used by volume bars */

	plVidType=vidNorm;
	plScrType=0;
	plScrMode=0;
	plActualMode=0;
	plLFB=0;
	RefreshScreen();

	Height=plScrHeight=LINES;
	if ((Width=plScrWidth=COLS)>CONSOLE_MAX_X)
		Width=plScrWidth=CONSOLE_MAX_X;
	else if (plScrWidth<80)
		Width=plScrWidth=80; /* If a console gets smaller than THIS, the user deserves to get fucked up his ASS */

	vgaMakePal();

	conRestore();

	return 0;
}

void curses_done(void)
{
	conRestore();
}
